require "chord_finder/note"
require "chord_finder/scale"
require "chord_finder/chord"
require "chord_finder/chord_position"
require "chord_finder/fretboard"
require "chord_finder/chart_printer"

class ChordFinder
  attr_reader :fretboard, :printer

  def initialize(tuning = :EADGBE, fret_count = 20)
    @fretboard = Fretboard.new(tuning, fret_count)
    @printer = ChartPrinter.new
  end

  def display(chord, options)
    chord = build_chord(chord)
    positions = find(chord, options)

    if options[:simple]
      positions.each { |position| puts position.to_s }
    else
      positions.each { |position| printer.call(position, fretboard.open_notes) }
    end

    "Notes: #{chord.notes.join(" ")}"
  end

  def find(chord, options)
    chord_frets = fretboard.filter_frets(chord, options)
    positions   = create_positions(chord_frets, chord)
    valid_positions = select_valid_positions(positions, chord)

    # prioritize positions with chord root in the bass
    first_in_bass, fifth_in_bass =
      valid_positions.partition { |pos| pos.bass_note == chord.first }

    (first_in_bass + fifth_in_bass).take(options[:limit] || 5)
  end

  private
  def build_chord(chord)
    case chord
    when String then Chord.new(chord)
    when Symbol then Chord.new(chord)
    when Chord  then chord
    end
  end

  def select_valid_positions(positions, chord)
    valid_positions = []
    positions.each do |position|
      next unless position.valid?
      i = valid_positions.find_index { |prev_position|
                            position.same_form?(prev_position) }

      if i
        prev_position = valid_positions[i]
        valid_positions[i] =
          position.num_sounding_strings > prev_position.num_sounding_strings ?
          position : prev_position
      else
        valid_positions << position
      end
    end
    valid_positions
  end

  def create_positions(chord_frets, chord)
    ChordPosition.create_list(chord_frets[0].product(*chord_frets[1..-1]), chord)
  end
end
