class ChordPosition
  def self.create_list(frets_list, chord)
    frets_list.map { |frets| ChordPosition.new(frets, chord) }
  end

  attr_reader :frets, :chord

  def initialize(frets, chord)
    @frets = frets
    @chord = chord
  end

  def to_s
    fret_nums.join.gsub("-1", "x")
  end

  def same_form?(other_position)
    same_bass?(other_position) &&
    same_on_sounding_strings?(other_position)
  end

  def valid?
    chord?                  &&
    frettable?              &&
    includes_bass_string?   &&
    first_or_fifth_in_bass?
  end

  def fret_nums
    @fret_nums ||= frets.map(&:number)
  end

  def num_sounding_strings
    6 - num_muted_strings
  end

  def num_muted_strings
    fret_nums.count(-1)
  end

  def closed_fret_nums
    @closed_fret_nums ||= fret_nums.select { |num| num > 0 }
  end

  def lowest_fret_num
    closed_fret_nums.min || 0
  end

  def highest_fret_num
    closed_fret_nums.max || 0
  end

  def notes
    @notes ||= frets.map(&:note).select { |note| note.to_s != "x" }
  end

  def bass_note
    frets.find { |fret| fret.note.to_s != "x" }.note
  end

  def bass_note_string
    frets.find_index { |fret| fret.note.to_s != "x" }
  end

  private
  def chord?
    notes.uniq.sort_by(&:position) == chord.notes.sort_by(&:position)
  end

  def frettable?
    closed_fret_nums.empty? || (highest_fret_num - lowest_fret_num) <= 3
  end

  def includes_bass_string?
    (0..2).include? bass_note_string
  end

  def first_or_fifth_in_bass?
    bass_note == chord.first || bass_note == chord.fifth
  end

  def same_bass?(other_position)
    s1, s2 = bass_note_string, other_position.bass_note_string
    s1 == s2 && frets[s1] == frets[s2]
  end

  def same_on_sounding_strings?(other_position)
    fret_nums.zip(other_position.fret_nums).all? { |num1, num2|
      (num1 >= 0 && num2 >= 0) ? num1 == num2 : true }
  end
end
